/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */
#include "config.h"

#include <X11/Intrinsic.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xatom.h>

#include <stdio.h>
#include <stdlib.h>
#include "xtiffio.h"
#include "geotiff.h"
#include "geotiffio.h"
#include "geo_normalize.h"
#include "cpl_csv.h"
#include "xmap.h"
#include "image.h"
#include "debug.h"


extern XtAppContext app;
extern Widget draww;

#define HASH_SIZE 128
#define HASH(c)           ((int)((c).red + (c).green + (c).blue) % HASH_SIZE)
#define HASH_PIXEL(c)     ((c) % HASH_SIZE)
typedef struct hash_s {
    int *val;
    struct hash_s *left, *right, *same, **parentp;
} HashElement;

typedef struct {
    int (*cmp) (void *, void *);
    void (*free) (void *);
    HashElement **table;
    int nelem;
} HashTable;

#define ZINDEX8(x, y, img)  ((y) * img->bytes_per_line) + (x)

typedef struct {
    int isGrey;
    int isDefault;
    int isMapped;
    int ncolors;
    int readonly;
    Display *display;
    Colormap cmap;
    Visual *visual;
    int depth;
    void *htable, *ltable;
    void *ctable;
    Pixel mine;
    /*
     **  TrueColor visual support
     */
    int rShift, gShift, bShift;
    int rRange, gRange, bRange;
    /*
     **  Free color entries
     */
    int nfree;
    void *list;
    /*
     **  Cache the last request value here
     */
    void *last;
} Palette;

typedef struct list_s {
    Display *dpy;
    Colormap cmap;
    Palette *map;
    struct list_s *next;
} PaletteList;


typedef struct Col_s {
    XColor color;
    Boolean used;
    Boolean invalid;
} Col;


static int forceRGB;
static unsigned short planarconfig;
static unsigned short *redmap, *greenmap, *bluemap;
static PaletteList *cmapList = NULL;
static TIFF *tif;
static GTIF *gtif;
static GTIFDefn defn;
static unsigned short bitsPS, samplesPP, photometric;
static int width, height, rows_per_strip, current_strip, strip_size;
static int maxval, ncolors, gtifstatus;
static char *strip_buf;

#define STEP    256


XImage *NewXImage(Display * dpy, Visual * visual, int depth, int width, int height)
{
    XImage *xim;
    int pad;

    if (depth > 16)
	pad = 32;
    else if (depth > 8)
	pad = 16;
    else
	pad = 8;

    xim = XCreateImage(dpy, visual, depth, ZPixmap, 0, NULL,
		       width, height, pad, 0);

    if (xim == NULL)
	return NULL;

    xim->data = (char *) XtMalloc(xim->bytes_per_line * height);

    if (xim->data == NULL) {
	XDestroyImage(xim);
	xim = NULL;
    }
    return xim;
}


Widget GetShell(Widget w)
{
    while (!XtIsShell(w))
	w = XtParent(w);
    return w;
}

void *HashFind(void *t, int value, void *val)
{
    HashTable *tbl = (HashTable *) t;
    HashElement *cur;
    int v;

    if (tbl == NULL)
	return NULL;

    cur = tbl->table[value];

    while (cur != NULL) {
	if ((v = tbl->cmp(cur->val, val)) == 0)
	    return cur->val;

	if (v < 0)
	    cur = cur->left;
	else
	    cur = cur->right;
    }

    return NULL;
}

static HashElement *find(HashElement * pos, void *elem)
{
    HashElement *v;

    if (pos == NULL)
	return NULL;

    if (pos->val == elem)
	return pos;
    if (pos->left != NULL && (v = find(pos->left, elem)) != NULL)
	return v;
    if (pos->right != NULL && (v = find(pos->right, elem)) != NULL)
	return v;
    if (pos->same != NULL && (v = find(pos->same, elem)) != NULL)
	return v;
    return NULL;
}


int HashRemove(void *t, int value, void *elem)
{
    HashTable *tbl = (HashTable *) t;
    HashElement *v;

    if (elem == NULL || tbl == NULL)
	return 1;

    if ((v = find(tbl->table[value], elem)) == NULL)
	return 0;		/* The element was not actually in the table */

    if (v->same != NULL) {
	/*
	   **  Same links are not allowed to have left & right children
	   **   so just inherit the parent's children.
	 */
	if (v->left != NULL)
	    v->left->parentp = &v->same->left;
	if (v->right != NULL)
	    v->right->parentp = &v->same->right;
	v->same->left = v->left;
	v->same->right = v->right;

	*v->parentp = v->same;
	v->same->parentp = v->parentp;
    } else if (v->left != NULL) {
	*v->parentp = v->left;
	v->left->parentp = v->parentp;
	if (v->right != NULL) {
	    HashElement *cur = v->left, **prev = &v->left;
	    while (cur != NULL) {
		if (tbl->cmp(cur->val, v->right->val) < 0) {
		    prev = &cur->left;
		    cur = cur->left;
		} else {
		    prev = &cur->right;
		    cur = cur->right;
		}
	    }
	    *prev = v->right;
	    v->right->parentp = prev;
	}
    } else {
	/* no left side, just attach the right */
	*v->parentp = v->right;
	if (v->right != NULL)
	    v->right->parentp = v->parentp;
    }
    free(v);

    return 1;
}


int HashAdd(void *t, int value, void *val)
{
    HashTable *tbl = (HashTable *) t;
    HashElement *new;
    HashElement *cur = tbl->table[value];
    HashElement **prev;
    int v;

    if (tbl == NULL)
	return 1;

    new = (HashElement *) malloc(sizeof(HashElement));

    new->left = NULL;
    new->right = NULL;
    new->same = NULL;
    new->val = val;
    new->parentp = NULL;

    prev = &tbl->table[value];
    while (cur != NULL) {
	if ((v = tbl->cmp(cur->val, val)) < 0) {
	    prev = &cur->left;
	    cur = cur->left;
	} else if (v > 0) {
	    prev = &cur->right;
	    cur = cur->right;
	} else {
	    prev = &cur->same;
	    new->same = cur->same;
	    if (cur->same != NULL)
		cur->same->parentp = &new->same;
	    break;
	}
    }
    *prev = new;
    new->parentp = prev;

    return 0;
}


void * HashCreate(int (*cmp) (void *, void *), void (*free) (void *), int nelem)
{
    HashTable *new;
    int i;

    new = (HashTable *) malloc(sizeof(HashTable));
    new->nelem = nelem;
    new->cmp = cmp;
    new->free = free;
    new->table = (HashElement **) malloc(nelem * sizeof(HashElement *));

    for (i = 0; i < nelem; i++)
	new->table[i] = NULL;

    return (void *) new;
}


static int readCMP(void *cca, void *ccb)
{
    XColor *ca = &((Col *) cca)->color, *cb = &((Col *) ccb)->color;

    if (ca->red != cb->red)
	return ca->red < cb->red ? -1 : 1;
    if (ca->green != cb->green)
	return ca->green < cb->green ? -1 : 1;
    if (ca->blue != cb->blue)
	return ca->blue < cb->blue ? -1 : 1;
    return 0;
}

static int cmpPixel(void *cca, void *ccb)
{
    if (((Col *) cca)->color.pixel == ((Col *) ccb)->color.pixel)
	return 0;
    return (((Col *) cca)->color.pixel < ((Col *) ccb)->color.pixel) ? -1 : 1;
}

static void freeFunc(void *junk)
{
}


static void entryUnlink(Palette * map, Col * node)
{
    node->used = True;
    node->invalid = False;
    map->nfree--;
}

static Palette *paletteNew(Widget w, Boolean useDefault)
{
    int i, v;
    Display *dpy = XtDisplay(w);
    Screen *screen = XtScreen(w);
    Colormap rcmap = DefaultColormapOfScreen(screen);
    Palette *map = XtNew(Palette);
    Col *ctable;
    Visual *visual = NULL;
    PaletteList *new;
    int end, depth = -1;
    Boolean flg = False;

    XtVaGetValues(w, XtNvisual, &visual, XtNdepth, &depth, NULL);
    if (visual == NULL || useDefault)
	visual = DefaultVisualOfScreen(screen);
    if (depth <= 0 || useDefault)
	depth = DefaultDepthOfScreen(screen);

    map->htable = NULL;
    map->ltable = NULL;
    map->display = dpy;
    map->htable = HashCreate(readCMP, freeFunc, HASH_SIZE);
    map->ltable = HashCreate(cmpPixel, freeFunc, HASH_SIZE);
    map->list = NULL;
    map->last = NULL;
    map->isMapped = visual->class != TrueColor;
    map->isGrey = (visual->class == StaticGray || visual->class == GrayScale);
    map->ncolors = visual->map_entries;
    map->nfree = 0;
    map->ctable = NULL;
    map->visual = visual;
    map->depth = depth;
    map->isDefault = False;
    map->mine = None;

    switch (visual->class) {
    case TrueColor:
	map->rShift = 0;
	map->gShift = 0;
	map->bShift = 0;
	map->rRange = 1;
	map->gRange = 1;
	map->bRange = 1;
	for (v = visual->red_mask; (v & 1) == 0; v >>= 1)
	    map->rShift++;
	for (; (v & 1) == 1; v >>= 1)
	    map->rRange <<= 1;
	for (v = visual->green_mask; (v & 1) == 0; v >>= 1)
	    map->gShift++;
	for (; (v & 1) == 1; v >>= 1)
	    map->gRange <<= 1;
	for (v = visual->blue_mask; (v & 1) == 0; v >>= 1)
	    map->bShift++;
	for (; (v & 1) == 1; v >>= 1)
	    map->bRange <<= 1;
    case StaticGray:
    case StaticColor:
	map->readonly = True;
	map->cmap = XCreateColormap(dpy, RootWindowOfScreen(screen),
				    visual, AllocNone);
	map->isDefault = False;
	goto addlist;
    default:
	map->readonly = False;
	if (useDefault) {
	    map->cmap = rcmap;
	    map->isDefault = True;
	} else {
	    map->cmap = XCreateColormap(dpy, RootWindowOfScreen(screen),
					visual, AllocAll);
	}
	break;
    }

    ctable = (Col *) XtCalloc(sizeof(Col), visual->map_entries);
    map->ctable = ctable;
    end = CellsOfScreen(screen);
    for (i = 0; i < visual->map_entries; i += STEP) {
	XColor xcol[STEP];
	int cnt = visual->map_entries - i;
	int j, d;

	if (cnt > STEP)
	    cnt = STEP;

	for (j = 0; j < cnt; j++) {
	    Col *c = &ctable[i + j];

	    c->color.pixel = i + j;
	    c->color.flags = DoRed | DoGreen | DoBlue;
	    xcol[j].pixel = i + j;
	    xcol[j].flags = DoRed | DoGreen | DoBlue;

	    c->used = False;
	    c->invalid = False;
	}

	if (i >= end)
	    d = 0;
	else
	    d = MIN(end - i, cnt);

	if (i < end)
	    XQueryColors(dpy, rcmap, xcol, d);
	if (!flg) {
	    for (j = d; j < cnt; j++) {
		xcol[j].flags = DoRed | DoGreen | DoBlue;
		xcol[j].red = xcol[j].green = xcol[j].blue = 0xffff;
	    }
	    if (d == 0)
		flg = True;
	}
	if (!map->isDefault)
	    XStoreColors(dpy, map->cmap, xcol, cnt);

	for (j = 0; j < cnt; j++) {
	    Col *c = &ctable[i + j];

	    c->color.red = xcol[j].red & 0xff00;
	    c->color.green = xcol[j].green & 0xff00;
	    c->color.blue = xcol[j].blue & 0xff00;
	    HashAdd(map->htable, HASH(c->color), c);
	    HashAdd(map->ltable, HASH_PIXEL(c->color.pixel), c);
	}
    }

    map->nfree = visual->map_entries;

    if (!map->isDefault) {
	Boolean got = False;

	got = !map->isMapped;
	for (i = 0; i < visual->map_entries; i++) {
	    if (ctable[i].color.pixel == BlackPixelOfScreen(screen)) {
		entryUnlink(map, &ctable[i]);
	    } else if (ctable[i].color.pixel == WhitePixelOfScreen(screen)) {
		entryUnlink(map, &ctable[i]);
	    } else if (!got) {
		map->mine = i;
		entryUnlink(map, &ctable[i]);
		got = True;
	    }
	}
    }
  addlist:
    new = (PaletteList *) XtMalloc(sizeof(PaletteList));
    new->dpy = XtDisplay(w);
    new->cmap = map->cmap;
    new->next = cmapList;
    new->map = map;
    cmapList = new;

    return map;
}

Palette *PaletteCreate(Widget w)
{
    return paletteNew(GetShell(w), False);
}


Image *ImageNew(int width, int height)
{
    Image *image = XtNew(Image);

    image->refCount = 1;
    image->isBW = 0;
    image->isGrey = 0;
    image->cmapPacked = 0;
    image->cmapSize = 0;
    image->cmapData = NULL;
    image->width = width;
    image->height = height;
    image->scale = 3;
    if (width == 0 || height == 0)
	image->data = NULL;
    else
	image->data = (unsigned char *) XtMalloc(width * height * sizeof(char) * 3);
    image->maskData = NULL;
    return image;
}

void ImageDelete(Image * image)
{
    image->refCount--;
    if (image->refCount > 0)
	return;

    if (image->cmapSize > 0 && image->cmapData != NULL)
	XtFree((XtPointer) image->cmapData);
    if (image->data != NULL)
	XtFree((XtPointer) image->data);
    if (image->maskData != NULL)
	XtFree((XtPointer) image->maskData);
    XtFree((XtPointer) image);
}


Image *ImageNewCmap(int width, int height, int size)
{
    Image *image = ImageNew(0, 0);

    if (size == 0)
	image->scale = 3;
    else if (size <= 256)
	image->scale = 1;
    else
	image->scale = sizeof(short) / sizeof(char);

    image->width = width;
    image->height = height;
    image->data = (unsigned char *)
	XtMalloc(width * height * sizeof(char) * image->scale);
    if (size != 0)
	image->cmapData = (unsigned char *) XtMalloc(size * sizeof(char) * 3);
    image->cmapSize = size;

    return image;
}


static void nop(void)
{
}

static int getImage(TIFF * tif, Image * image, void *dp,
		    unsigned short bitsPS,
		    unsigned short samplesPP,
		    unsigned short photometric,
		    int zoom)
{
    unsigned short orientation;
    int row, rowStart, colStart, rowInc, colInc, ret;
    int isWide;
    unsigned char *ucp;
    unsigned short *usp;
    unsigned char *sline = NULL, *sptr0, *sptr1, *sptr2;
    int x, y;
    int maxval = (1 << bitsPS) - 1;
    int scansize = 0;

    if (planarconfig == PLANARCONFIG_SEPARATE) {
	scansize = TIFFStripSize(tif);

	sline = (unsigned char *) malloc(3 * scansize);
	sptr0 = sline;
	sptr1 = sptr0 + scansize;
	sptr2 = sptr1 + scansize;
    } else {
	scansize = TIFFScanlineSize(tif);
	if ((sline = (unsigned char *) malloc(scansize)) == NULL) {
	    error("Unable to allocate memory for scanline.", 0);
	    return 1;
	}
    }

    isWide = image->cmapSize > 256;

    usp = (unsigned short *) dp;
    ucp = (unsigned char *) dp;

    if (!TIFFGetField(tif, TIFFTAG_ORIENTATION, &orientation))
	orientation = ORIENTATION_TOPLEFT;

    switch (orientation) {
    default:
    case ORIENTATION_LEFTTOP:
    case ORIENTATION_TOPLEFT:
	colStart = 0;
	colInc = 1;
	rowStart = image->firstline;
	rowInc = 1;
	break;
    case ORIENTATION_RIGHTTOP:
    case ORIENTATION_TOPRIGHT:
	colStart = image->width;
	colInc = -1;
	rowStart = image->firstline;
	rowInc = 1;
	break;
    case ORIENTATION_RIGHTBOT:
    case ORIENTATION_BOTRIGHT:
	colStart = image->width;
	colInc = -1;
	rowStart = image->height - image->firstline;
	rowInc = -1;
	break;
    case ORIENTATION_LEFTBOT:
    case ORIENTATION_BOTLEFT:
	colStart = 0;
	colInc = 1;
	rowStart = image->height - image->firstline;
	rowInc = -1;
	break;
    }

    rowInc *= zoom;

    for (row = rowStart, y = image->firstline; y < image->lastline; y++, row += rowInc) {
	unsigned char *pp = sline;
	int bitsleft, sample0, sample1, sample2;

#define NEXT3(p0, p1, p2) {			\
	if (bitsleft == 0) {			\
	  p0++;					\
	  p1++;					\
	  p2++;					\
	  bitsleft = 8;				\
	}					\
	bitsleft -= bitsPS;			\
	sample0 = (*p0 >> bitsleft) & maxval;	\
	sample1 = (*p1 >> bitsleft) & maxval;	\
	sample2 = (*p2 >> bitsleft) & maxval;	\
}

#define NEXT(pp) do {				\
	if (bitsleft == 0) {			\
	  pp++;					\
	  bitsleft = 8;				\
	}					\
	bitsleft -= bitsPS;			\
	sample0 = (*pp >> bitsleft) & maxval;	\
} while (0)

	if (planarconfig == PLANARCONFIG_SEPARATE) {
	    if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 0),
				     sptr0, scansize) < 0)
		break;
	    if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 1),
				     sptr1, scansize) < 0)
		break;
	    if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 2),
				     sptr2, scansize) < 0)
		break;
	} else if (rows_per_strip == 1) {
	    if (TIFFReadScanline(tif, sline, row, 0) < 0)
		break;
	} else {
	    if (current_strip != TIFFComputeStrip(tif, row, 0)) {
		if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 0),
                                     strip_buf, strip_size) < 0)
		    break;
		current_strip = TIFFComputeStrip(tif, row, 0);
	    }
	    memcpy(sline, strip_buf + (scansize * (row%rows_per_strip)),
		scansize); 
	}


	/*
	 **  If colInc == -1, then just reorder the samples.  
	 */

	bitsleft = 8;

	switch (photometric) {
	case PHOTOMETRIC_PALETTE:
	    if (forceRGB) {

		for (x = 0; x < image->width; x++) {
		    NEXT(pp);
		    *ucp++ = redmap[sample0];
		    *ucp++ = greenmap[sample0];
		    *ucp++ = bluemap[sample0];
		}
		break;
	    }
	    /*  fall through        */
	case PHOTOMETRIC_MINISWHITE:
	case PHOTOMETRIC_MINISBLACK:
	    for (x = 0; x < image->width; x++) {
		NEXT(pp);
		if (isWide)
		    *usp++ = sample0;
		else
		    *ucp++ = sample0;
	    }
	    break;
	case PHOTOMETRIC_RGB:
	    if (planarconfig == PLANARCONFIG_SEPARATE) {
		unsigned char *p0 = sptr0, *p1 = sptr1, *p2 = sptr2;
		for (x = 0; x < image->width; x++) {
		    NEXT3(p0, p1, p2);
		    *ucp++ = sample0;
		    *ucp++ = sample1;
		    *ucp++ = sample2;
		}
	    } else {
		for (x = 0; x < image->width; x++) {
		    NEXT(pp);
		    *ucp++ = sample0;
		    NEXT(pp);
		    *ucp++ = sample0;
		    NEXT(pp);
		    *ucp++ = sample0;
		    if (samplesPP == 4)
			NEXT(pp);
		}
	    }
	    break;
	case PHOTOMETRIC_MASK:
	    for (x = 0; x < image->width; x++) {
		NEXT(pp);
		*ucp++ = sample0 ? 1 : 0;
	    }
	    break;
	}
    }
    if (sline != NULL)
	free(sline);

    return 0;
}

Image *ReadTIFF(char *file, int first, int last, int zoom)
{
    Image *image = NULL;
    int i;

    forceRGB = 0;

    last = MIN(last, height);

    if (maxval == 1 && samplesPP == 1) {
	image = ImageNewCmap(width, last - first, ncolors);
	if (photometric == PHOTOMETRIC_MINISBLACK) {
	    ImageSetCmap(image, 0, 0, 0, 0);
	    ImageSetCmap(image, 1, 255, 255, 255);
	} else {
	    ImageSetCmap(image, 1, 0, 0, 0);
	    ImageSetCmap(image, 0, 255, 255, 255);
	}
    } else {
	switch (photometric) {
	case PHOTOMETRIC_MINISBLACK:
	    image = ImageNewCmap(width, last - first, ncolors);
	    for (i = 0; i < ncolors; i++) {
		int v = (i * 255) / maxval;
		ImageSetCmap(image, i, v, v, v);
	    }
	    break;
	case PHOTOMETRIC_MINISWHITE:
	    image = ImageNewCmap(width, last - first, ncolors);
	    for (i = 0; i < ncolors; i++) {
		int v = ((maxval - i) * 255) / maxval;
		ImageSetCmap(image, i, v, v, v);
	    }
	    break;

	case PHOTOMETRIC_PALETTE:
	    if (!TIFFGetField(tif, TIFFTAG_COLORMAP, &redmap, &greenmap, &bluemap)) {
		error("error getting colormaps.", 0);
		return NULL;
	    }
	    if (ncolors > 0xffff) {
		forceRGB = 1;
		image = ImageNew(width, last - first);
		break;
	    } else {
		image = ImageNewCmap(width, last - first, ncolors);
	    }
	    for (i = 0; i < ncolors; i++) {
		ImageSetCmap(image, i, redmap[i] / 256, greenmap[i] / 256, bluemap[i] / 256);
	    }
	    break;

	case PHOTOMETRIC_RGB:
	    image = ImageNew(width, last - first);
	    break;
	}
    }

    image->firstline = first;
    image->lastline = last;

    if (getImage(tif, image, (void *) image->data, bitsPS, samplesPP, photometric, zoom)) {
	ImageDelete(image);
	image = NULL;
    }
    return image;
}

static void addColor(Palette * map, XColor * color)
{
    Col *node, *n, *cptr;
    int i;


    if (map->readonly) {
	/*
	 **  The temporary color is needed since alloc
	 **  will change the pixel values.
	 */
	XColor newc;
	newc = *color;
	node = XtNew(Col);

	newc.flags = DoRed | DoGreen | DoBlue;
	myXAllocColor(map->display, map->cmap, &newc);
	node->color.pixel = color->pixel = newc.pixel;
    } else {
	if (map->nfree == 0) {
	    unsigned int d, curDif = ~0;
	    int rd, gd, bd;
	    /*
	     **  All free colors used up -- use closest match.
	     */
	    for (i = 0, cptr = (Col *) map->ctable; i < map->ncolors; i++, cptr++) {
		rd = (cptr->color.red >> 8) - (color->red >> 8);
		gd = (cptr->color.green >> 8) - (color->green >> 8);
		bd = (cptr->color.blue >> 8) - (color->blue >> 8);
		d = rd * rd + gd * gd + bd * bd;
		if (d < curDif) {
		    node = cptr;
		    curDif = d;
		}
	    }
	    color->pixel = node->color.pixel;
	    return;
	}
	/* Find first free entry -- we know it's there. */
	for (i = 0, node = (Col *) map->ctable;
	     (i < map->ncolors) && node->used; i++, node++);
	entryUnlink(map, node);	/* mark as used */

	color->pixel = node->color.pixel;
	n = HashFind(map->ltable, HASH_PIXEL(node->color.pixel), node);
	HashRemove(map->ltable, HASH_PIXEL(node->color.pixel), n);
	HashRemove(map->htable, HASH(*color), node);
    }

    node->used = True;
    node->invalid = False;
    node->color.red = color->red & 0xff00;
    node->color.green = color->green & 0xff00;
    node->color.blue = color->blue & 0xff00;
    node->color.flags = DoRed | DoGreen | DoBlue;
    if (!map->readonly)
	XStoreColor(map->display, map->cmap, &node->color);
    HashAdd(map->htable, HASH(*color), node);
    HashAdd(map->ltable, HASH_PIXEL(color->pixel), node);
}


/*
 * Try to allocate the specified XColor in the colormap.
 * If the color is already in the palette, use that instead.
 * Return the corresponding Pixel value in 'list'.
 */
static Col *allocN(Palette * map, XColor * color, Pixel * list)
{
    Col c, *node;

    c.color.red = color->red & 0xff00;
    c.color.green = color->green & 0xff00;
    c.color.blue = color->blue & 0xff00;

    if ((node = HashFind(map->htable, HASH(c.color), &c)) == NULL) {
	addColor(map, color);
	*list = color->pixel;
    } else {
	/*
	 **  It must have been allocated in the previous
	 **   pass, or by the above.
	 */
	if (!node->used)
	    entryUnlink(map, node);
	*list = node->color.pixel;
    }

    return node;
}

/*
 * Return the Pixel corresponding to (or closest to) the RGB
 * values specified in 'color'.
 * If no match is found in the existing palette, try to allocate
 * a new color.
 */
Pixel PaletteAlloc(Palette * map, XColor * color)
{
    if (!map->isMapped) {
	unsigned int r, g, b;

	r = (color->red * map->rRange) >> 16;
	g = (color->green * map->gRange) >> 16;
	b = (color->blue * map->bRange) >> 16;
	return (r << map->rShift) | (g << map->gShift) | (b << map->bShift);
    }
    if (map->last != NULL) {
	XColor *lc = (XColor *) map->last;

	if (lc->red == color->red &&
	    lc->green == color->green &&
	    lc->blue == color->blue)
	    return color->pixel = lc->pixel;
    }
    map->last = allocN(map, color, &color->pixel);

    return color->pixel;
}

//#define KEEPXALIVE() {if (y % step == 0) XtAppProcessEvent(app, XtIMAll);}
#define KEEPXALIVE()

static void imageToPixmapLoop(Display * dpy, Image * image,
		       Palette * map, XImage * xim, Pixmap pix, int zoom)
{
    GC gc;
    int x, y;
    int width = image->width, height = image->height;
    Colormap cmap = DefaultColormapOfScreen(XtScreen(draww));
    int step = 64 * 256 / width;

    if (image->cmapSize > 0) {
	unsigned short *sdp = (unsigned short *) image->data;
	unsigned char *cdp = (unsigned char *) image->data;
	Pixel *list = (Pixel *) XtCalloc(sizeof(Pixel), image->cmapSize);
	XColor *xcol = (XColor *) XtCalloc(sizeof(XColor), image->cmapSize);

	for (y = 0; y < image->cmapSize; y++) {
	    xcol[y].red = image->cmapData[y * 3 + 0] << 8;
	    xcol[y].green = image->cmapData[y * 3 + 1] << 8;
	    xcol[y].blue = image->cmapData[y * 3 + 2] << 8;
	    myXAllocColor(map->display, cmap, &xcol[y]);
	    list[y] = xcol[y].pixel;
	}

#if 1
	if (xim->bits_per_pixel == 8) {
	    unsigned char *data = (unsigned char *) xim->data;

	    if (image->cmapSize > 256) {
		for (y = 0; y < height; y++) {
		    sdp = ((unsigned short *) image->data) + y * width;
		    for (x = 0; x < width / zoom; x++, sdp += zoom)
			data[ZINDEX8(x, y, xim)] = list[*sdp];
		    KEEPXALIVE();
		}
	    } else {
		for (y = 0; y < height; y++) {
		    cdp = ((unsigned char *) image->data) + y * width;
		    for (x = 0; x < width / zoom; x++, cdp += zoom)
			data[ZINDEX8(x, y, xim)] = list[*cdp];
		    KEEPXALIVE();
		}
	    }
	} else
#endif
#if 1
	if (xim->bits_per_pixel == 16 && !(image->cmapSize > 256)) {
	    /*
	     ** Optimized for 16 bit visuals and cmaps<=256 
	     */
	    unsigned char *addr;
	    unsigned long pixel;

	    if (xim->byte_order == MSBFirst) {
		for (y = 0; y < height; y++) {
		    cdp = ((unsigned char *) image->data) + y * width;
		    for (x = 0; x < width / zoom; x++, cdp += zoom) {
			pixel = list[*cdp];
			addr = &((unsigned char *) xim->data)
			    [y * xim->bytes_per_line + (x << 1)];
			addr[0] = pixel >> 8;
			addr[1] = pixel;
		    }
		    KEEPXALIVE();
		}
	    } else {
		for (y = 0; y < height; y++) {
		    cdp = ((unsigned char *) image->data) + y * width;
		    for (x = 0; x < width / zoom; x++, cdp += zoom) {
			pixel = list[*cdp];
			addr = &((unsigned char *) xim->data)
			    [y * xim->bytes_per_line + (x << 1)];
			addr[1] = pixel >> 8;
			addr[0] = pixel;
		    }
		    KEEPXALIVE();
		}
	    }
	} else
#endif
	{
	    /*
	     **  Slow loop
	     */
	    if (image->cmapSize > 256) {
		for (y = 0; y < height; y++) {
		    sdp = ((unsigned short *) image->data) + y * width;
		    for (x = 0; x < width / zoom; x++, sdp += zoom)
			XPutPixel(xim, x, y, list[*sdp]);
		}
	    } else {
		for (y = 0; y < height; y++) {
		    cdp = ((unsigned char *) image->data) + y * width;
		    for (x = 0; x < width / zoom; x++, cdp += zoom)
			XPutPixel(xim, x, y, list[*cdp]);
		}
	    }
	}

	XtFree((XtPointer) list);
	XtFree((XtPointer) xcol);
    } else {
	unsigned char *cp = image->data;

	for (y = 0; y < height; y++) {
	    for (x = 0; x < width; x++) {
		XColor c;
		Pixel p;

		c.red = *cp++ << 8;
		c.green = *cp++ << 8;
		c.blue = *cp++ << 8;

		p = PaletteAlloc(map, &c);

		if (xim->bits_per_pixel == 8)
		    xim->data[ZINDEX8(x, y, xim)] = p;
		else
		    XPutPixel(xim, x, y, p);
	    }
	    KEEPXALIVE();
	}
    }

    gc = XCreateGC(dpy, pix, 0, 0);
    XPutImage(dpy, pix, gc, xim, mapdata.mx_offset / mapdata.zoom, 0, 0, 0, width, height);
    XFreeGC(dpy, gc);

    XDestroyImage(xim);
    ImageDelete(image);
}

#define ImagePixel(image, x, y)                                         \
        (((image)->cmapSize == 0)                                       \
          ? &((image)->data[((y) * (image)->width + (x)) * 3])          \
          : (((image)->cmapSize > 256)                                  \
             ? &((image)->cmapData[((unsigned short *)(image)->data)    \
                                [(y) * (image)->width + (x)] * 3])      \
             : &((image)->cmapData[(image)->data[(y)*(image)->width+(x)] * 3])))


static void compressColormap(Image * image)
{
    unsigned char used[32768];
    int size = image->width * image->height;
    int i, count, newSize;
    unsigned char *newMap;
    unsigned short *isp;
    unsigned char *icp;

    if (image->cmapSize <= 2 || image->cmapPacked)
	return;

    memset(used, False, sizeof(used));

    /*
     **  Find out usage information over the colormap
     */
    count = 0;
    if (image->cmapSize > 256) {
	isp = (unsigned short *) image->data;
	for (i = 0; i < size && count != image->cmapSize; i++, isp++) {
	    if (!used[*isp]) {
		used[*isp] = True;
		count++;
	    }
	}
    } else {
	icp = image->data;
	for (i = 0; i < size && count != image->cmapSize; i++, icp++) {
	    if (!used[*icp]) {
		used[*icp] = True;
		count++;
	    }
	}
    }

    /*
     **  Colormap is fully used.
     */
    if (count == image->cmapSize) {
	image->cmapPacked = True;
	return;
    }
    /*
     **  Now build the remapping colormap, and
     **    set the index.
     */
    newSize = count;
    newMap = (unsigned char *) XtCalloc(count, sizeof(unsigned char) * 3);
    for (count = i = 0; i < image->cmapSize; i++) {
	if (!used[i])
	    continue;
	newMap[count * 3 + 0] = image->cmapData[i * 3 + 0];
	newMap[count * 3 + 1] = image->cmapData[i * 3 + 1];
	newMap[count * 3 + 2] = image->cmapData[i * 3 + 2];
	used[i] = count++;
    }

    if (image->cmapSize > 256 && newSize > 256) {
	isp = (unsigned short *) image->data;
	for (i = 0; i < size; i++, isp++)
	    *isp = used[*isp];
    } else if (image->cmapSize > 256 && newSize <= 256) {
	/*
	   **  Map a big colormap down to a small one.
	 */
	isp = (unsigned short *) image->data;
	icp = image->data;
	for (i = 0; i < size; i++, icp++, isp++)
	    *icp = used[*isp];

	image->data = (unsigned char *)
	    XtRealloc((XtPointer) image->data, size * sizeof(unsigned char));
    } else {
	icp = image->data;
	for (i = 0; i < size; i++, icp++)
	    *icp = used[*icp];
    }

    XtFree((XtPointer) image->cmapData);
    image->cmapData = newMap;
    image->cmapSize = newSize;
    image->cmapPacked = True;
    image->isGrey = False;
}

static Image *quantizeColormap(Image * input, Palette * map, Boolean flag)
{
    Image *output;
    unsigned char *op;
    int x, y;
    int ncol;

    if (!map->isGrey)
	return input;

    ncol = map->ncolors > 256 ? 256 : map->ncolors;
    output = ImageNewCmap(input->width, input->height, ncol);

    op = output->data;

    for (y = 0; y < ncol; y++) {
	unsigned char v = ((float) y / (float) (ncol - 1)) * 255.0;
	ImageSetCmap(output, y, v, v, v);
    }

    for (y = 0; y < input->height; y++) {
	for (x = 0; x < input->width; x++, op++) {
	    unsigned char *dp, v;

	    dp = ImagePixel(input, x, y);
	    v = (dp[0] * 11 + dp[1] * 16 + dp[2] * 5) >> 5;	/* pp=.33R+.5G+.17B */
	    *op = (int) (((float) v / 256.0) * (float) ncol);
	}
    }

    output->maskData = input->maskData;
    input->maskData = NULL;
    ImageDelete(input);
    return output;
}


Boolean ImageToPixmap(Image * image, Widget w, Pixmap pix, Colormap * cmap, int zoom)
{
    Display *dpy = XtDisplay(w);
    Palette *colmap;
    XImage *xim;
    int width = image->width, height = image->height;

    colmap = PaletteCreate(w);
    *cmap = colmap->cmap;

    if ((xim = NewXImage(dpy, NULL, colmap->depth, width, height)) == NULL) {
	return False;
    }
#if 1
    if (options.nicecol) {
	if ((image->cmapSize > colmap->ncolors) ||
	    (image->cmapSize == 0 && colmap->isMapped)) {
	    image = quantizeColormap(image, colmap, False);
	}
	if (image->cmapSize > 0)
	    compressColormap(image);
    }
#endif

    imageToPixmapLoop(dpy, image, colmap, xim, pix, zoom);

    return True;
}

void tiff_cleanup()
{
//    fprintf(stderr,"tiff_cleanup() called.\n");

    if (current_strip >= 0 && strip_buf != NULL) {
	free(strip_buf);
	strip_buf = NULL;
    }
    gtifstatus = 0;
    GTIFFree(gtif);
    TIFFClose(tif);
}

void gtif_project(double lat, double lon, Position * x, Position * y)
{
  if (gtifstatus == 1) {
        double xx, yy;

      xx = lon; 
      yy = lat; 
      if ((defn.Model == ModelTypeGeographic) || 
	  (GTIFProj4FromLatLong(&defn, 1, &xx, &yy) == TRUE)) {
        if( GTIFPCSToImage( gtif, &xx, &yy )) {
	  *x = ((int)xx - mapdata.mx_offset) / mapdata.zoom;
	  *y = ((int)yy - mapdata.my_offset) / mapdata.zoom;
          return;
        }
      }
  }
  *x = lon2x(lon);
  *y = lat2y(lat);
  return;
}

static int pcs_to_deg(GTIF *gtif, GTIFDefn *defn, double *dx, double *dy) 
{
  /* Try to transform the coordinate into PCS space */
  if( !GTIFImageToPCS( gtif, dx, dy ) )
    return FALSE;
  if( defn->Model == ModelTypeGeographic )
      return TRUE;
  if( GTIFProj4ToLatLong( defn, 1, dx, dy ) )
    return TRUE;
  return FALSE;
}

void gtif_decode(Position x, Position y, double *lat, double *lon)
{
    if (gtifstatus == 1) {
        x = x * mapdata.zoom + mapdata.mx_offset;
        y = y * mapdata.zoom + mapdata.my_offset;
      *lat = y;
      *lon = x;
      if (pcs_to_deg(gtif, &defn, lon, lat) == TRUE) 
        return;
    }
    *lon = x2lon(x);
    *lat = y2lat(y);
}

static const char *tiff_csv_file( const char *basename )
{
  static char csvPath[1024];
  char dirbuf[1024];

  GetPrivateProfileString("General", "csvdir", "csv", dirbuf, 1023, inifile);
  snprintf( csvPath, 1023, "%s/%s", dirbuf, basename );
  
  return( csvPath );
}

int open_tiff_map(struct map *map)
{
    int i,geovers[3],geokeys;
    char bufp[256];
    char *lbuf;
    unsigned short count;
    double *data;
    extern Widget status_1;
    TIFF *tif2;
  
    gtifstatus = 0;
    geokeys = 0;
    forceRGB = 0;

    snprintf(bufp, 255, "%s/%s", get_mapdir(), map->filename);
    TIFFSetErrorHandler((TIFFErrorHandler) nop);

    if ((tif2 = XTIFFOpen(bufp, "r")) == NULL) {
	error("Error opening input file.", 0);
	return 0;
    }
    tif = tif2;

    if ((gtif = GTIFNew(tif)) == NULL) {
	error("Error opening GTIF Key parser.", 0);
	TIFFClose(tif);
	return 0;
    }

    SetCSVFilenameHook(tiff_csv_file);
    if (TIFFIsTiled(tif)) {
	error("Unable to handle TIFF tiled images.", 0);
	goto out;
    }

    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height);

    current_strip = -1;
    if (!TIFFGetField(tif, TIFFTAG_ROWSPERSTRIP, &rows_per_strip))
	rows_per_strip = 1;

    if (rows_per_strip > 1) {
	strip_size = TIFFStripSize(tif);
	strip_buf = malloc(strip_size);
    }

    map->mwidth = width;
    map->mheight = height;

    if (!TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitsPS))
	bitsPS = 1;

    if (!TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesPP))
	samplesPP = 1;

    if (!TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric)) {
	error("error getting photometric.", 0);
	goto out;
    }

    if (!strcmp(map->description,""))
      if (TIFFGetField(tif, TIFFTAG_IMAGEDESCRIPTION, &lbuf)) {
	free(map->description);
	map->description = strdup(lbuf);
	ch_title(map->description);
      }

      if( GTIFGetDefn( gtif, &defn ))
      {
	GTIFDefn *defnp = &defn;
        double x, y;
        mapdata.mx1 = 0;
        mapdata.my1 = 0;
      
        mapdata.mx2 = map->mwidth /*- 1*/;
        mapdata.my2 = map->mheight /*- 1*/;

	/* special case NAD27 */
	if (defnp->Datum == 6267) 
	  mapdata.mapdatum = 61;

	/* special case NAD83 */
	if (defnp->Datum == 6269) 
	  mapdata.mapdatum = 66;

        x = y = 0;
        /* If we can get this information, calibrate the map from it. */
        if (pcs_to_deg(gtif, &defn, &x, &y)) 
          {
            mapdata.lat1 = y;
            mapdata.lon1 = x;

            x = mapdata.mx2;
            y = mapdata.my2;
            if (pcs_to_deg(gtif, &defn, &x, &y)) 
              {
                mapdata.lat2 = y;
                mapdata.lon2 = x;
              }
  
            XmTextFieldSetString(status_1, 
                                 "Geotiff file geo-referenced map.");
            do_calibrate();
            gtifstatus = 1;
          }
    }    

    if (TIFFIsTiled(tif)) {
	error("Unable to handle TIFF tiled images.", 0);
	goto out;
    }

    if (samplesPP != 1 && samplesPP != 3 && samplesPP != 4) {
	error("Can only handle 1-channel gray scale\nor 1- or 3-channel color.", 0);
	goto out;
    }

    if (photometric == PHOTOMETRIC_YCBCR || photometric == PHOTOMETRIC_CIELAB ||
	photometric == PHOTOMETRIC_SEPARATED || photometric == PHOTOMETRIC_MASK) {
	error("Can only handle RGB, grey, or colormaped images.", 0);
	goto out;
    }

    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height);

    TIFFGetFieldDefaulted(tif, TIFFTAG_PLANARCONFIG, &planarconfig);

    if (planarconfig == PLANARCONFIG_SEPARATE && photometric != PHOTOMETRIC_RGB) {
	error("Can only handle planarconfig=SEPARATE on photometric=RGB images.", 0);
	goto out;
    }

    ncolors = 1 << bitsPS;
    maxval = ncolors - 1;

    register_cleanup(tiff_cleanup);

    // Now check the geo referencing...
    // Manual calibration takes precedence.
    //
    if (!mapdata.calibrated) {
	FILE *fp;

	i = strlen(map->filename);
	if (i>=4 && !strcasecmp(map->filename + (i-4), ".tif")) {
	    snprintf(bufp, 255, "%s/%s", get_mapdir(), map->filename);
	    if (map->filename[(i-3)] == 'T') {
	      i = strlen(bufp);
	      strcpy(bufp + (i-4), ".TFW");
	    } else {
	      i = strlen(bufp);
	      strcpy(bufp + (i-4), ".tfw");
	    }
	    if ((fp=fopen(bufp,"r")) != NULL) {
		geo_ref_tfw(fp, map);
		fclose(fp);
	    }
	}
    }

    return 1;

  out:
    tiff_cleanup();
    return 0;
}



get_tiff_map(struct map * map, Dimension xoff, Dimension yoff,
	     Dimension w, Dimension h, short zoom)
{
    Image *image;
    Colormap cmap;
    extern Widget draww;
    char bufp[256];

    snprintf(bufp, 255, "%s/%s", get_mapdir(), map->filename);
    XSync(XtDisplay(draww), False);

    if (image = ReadTIFF(bufp, yoff, yoff + h, zoom))
	ImageToPixmap(image, draww, map->pmap, &cmap, zoom);
}
